InstanceHub = class(Turbine.UI.Control)
function InstanceHub:Constructor()
    Turbine.UI.Control.Constructor(self)
    self.InstanceSwitch = function(sender, event, args)
        if type(event)=="function" then
            event(args)
        else
            if type(event)=="table" then
                for i = 1, #event do
                    if type(event[i])=="function" then
                        event[i](args)
                    end
                end
            end
        end
    end

    self.characterInfo = LoadTable(Turbine.DataScope.Character, "PrimeUtilsInstances") or {
        EttenmoorsState = false,
        InstanceVersion = 0,
        InstanceState = false,
        InstanceTier = 0,
        DelvingTier = 0,
        MissionQuest = "",
        DelvingMissionActive = false,
        MissionStartTime = 0,

    }

    self.TEQStats = LoadTable(Turbine.DataScope.Character, "PrimeUtilsTEQData") or  {

    }

    self:CreateTables()

    if Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment() == 2 then
        self.characterInfo.EttenmoorsState = true
        self.characterInfo.InstanceVersion = 4
    end

    if self.characterInfo.EttenmoorsState and Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment() < 2 then
        self.characterInfo.InstanceVersion = 5
    end
    
    self:ExecuteInstanceSwitch()

    if self.characterInfo.DelvingTier > 0 and self.characterInfo.MissionQuest ~= "" and self.characterInfo.DelvingMissionActive then
        self:DelvingHandling(true)
    end

    self.questCompleteCallback = AddCallback(ChatEventHandler, "QuestOrDeedCompleted", function(args)
        if self.characterInfo.DelvingMissionActive and string.match(args.Name, "Mission: (.*)") == self.characterInfo.MissionQuest then
            self:UpdateDelvingData(true)
            self:DelvingHandling(false)
        end
    end)

    self.chatCallback = AddCallback(ChatEventHandler, "QuestText", function(message) self:CheckInstanceState(message.Line) end)
end

function InstanceHub:EnterNewInstance(line, id, tier)
    self.characterInfo.InstanceState = true
    self.characterInfo.InstanceVersion = id
    if tier then
        self.characterInfo.InstanceTier = tier
    elseif line then
        line = string.gsub(line, "+", "")
        self.characterInfo.InstanceTier = string.sub(line, string.len(line), string.len(line))
    else
        self.characterInfo.InstanceTier = "?"
    end

    self:ExecuteInstanceSwitch()
end

function InstanceHub:GetInitialData()
    return {
        ["State"] = self.characterInfo.InstanceState, 
        ["Instance_Name"] = self.nameTable[self.characterInfo.InstanceVersion], 
        ["Instance_Tier"] = self.characterInfo.InstanceTier
    }
end

function InstanceHub:ExecuteInstanceSwitch()    
    self.InstanceSwitch(self, self.InstanceEvent, {["PvMP_State"] = self.characterInfo.EttenmoorsState, ["State"] = self.characterInfo.InstanceState, ["Instance_Name"] = self.nameTable[self.characterInfo.InstanceVersion], ["Instance_Tier"] = self.characterInfo.InstanceTier})
end

function InstanceHub:DelvingHandling(active)
    if active then
        --if not(self.characterInfo.DelvingMissionActive) then
            self.characterInfo.MissionStartTime = Turbine.Engine.GetGameTime()
        --end
        self.characterInfo.DelvingMissionActive = true
        self.characterInfo.InstanceState = true
        self.InstanceSwitch(self, self.DelvingEvent, {["DelvingMissionActive"] = self.characterInfo.DelvingMissionActive, ["Mission_Name"] = self.characterInfo.MissionQuest, ["Delving_Tier"] = self.characterInfo.DelvingTier})
    else
        self.InstanceSwitch(self, self.DelvingEvent, {["DelvingMissionActive"] = self.characterInfo.DelvingMissionActive, ["Mission_Name"] = self.characterInfo.MissionQuest, ["Delving_Tier"] = self.characterInfo.DelvingTier})
        self.characterInfo.DelvingTier = 0
        self.characterInfo.DelvingMissionActive = false
        self.characterInfo.InstanceState = false       
    end
end

function InstanceHub:CheckInstanceState(line)
    if string.find(line, "malice spreads through the area") then
        if self.characterInfo.DelvingTier > 0 and self.characterInfo.MissionQuest ~= "" then
            self:DelvingHandling(true)
        end
    end

    if string.find(line, "Failed: Tier") then
        self:UpdateDelvingData(false)
        self:DelvingHandling(false)
    end

    if string.find(line, "Quest") then
        local strippedLine = ""
        if string.find(line, "New Quest: Mission:") then
            if self.delvingMissions[line] then
                self.characterInfo.MissionQuest = self.delvingMissionsNames[self.delvingMissions[line]]
            end
        elseif string.find(line, ": Delving") then
            -- New Quest: Tier 3: Delving Bloodstone
            --Failed: Tier 3: Delving Bloodstone
            -- A formidable malice spreads through the area... 5
            self.characterInfo.DelvingTier = tonumber(string.match(line, "%d+"))
        elseif string.find(line, "Tier") then
            if string.find(line, "- ") then
                strippedLine = string.sub(line, 1, string.find(line, "- ") - 2)
                if self.questLines[strippedLine] then self:EnterNewInstance(line, self.questLines[strippedLine]) end
            end

            if string.match(line, "(.*) %(Tier [1-5]%)") then
                local instance, tier = string.match(string.gsub(line, "+", ""), "(.*) %(Tier (.*)%)")
                if self.questLines[instance] then self:EnterNewInstance(instance, self.questLines[instance], tier) end
            end


        elseif string.find(line, "Solo") or string.find(line, "Duo") then
            local instance = string.match(line, "(.*)%-%- Solo") or string.match(line, "(.*)%-%- Duo")

            if instance then
                if self.questLines[instance] then self:EnterNewInstance(instance, self.questLines[instance], 0) end
            end
        end

        if self.questLines[line] then 
            self:EnterNewInstance(line, self.questLines[line], "?") 
        end
    end

    if string.find(line, "Entered the") then
        if self.characterInfo.EttenmoorsState and not(string.find(line, "Ettenmoors" )) then
            self.characterInfo.EttenmoorsState = false
        end

        if string.find(string.match(line, "Entered the (.*) channel."), "-") and self.characterInfo.InstanceVersion > 0 then
            self.characterInfo.InstanceState = false
            self.characterInfo.InstanceVersion = 0
            self.characterInfo.InstanceTier = 0
            self:ExecuteInstanceSwitch()
            self.mplus:LoadDungeon(0, 0)
            

            self.InstanceSwitch(self, self.InstanceEvent, {["PvMP_State"] = self.characterInfo.EttenmoorsState, ["State"] = self.characterInfo.InstanceState, ["Instance_Name"] = self.nameTable[self.characterInfo.InstanceVersion], ["Instance_Tier"] = self.characterInfo.InstanceTier})
        end
    end

    if string.find(line, "Entered the Ettenmoors") and not(self.characterInfo.EttenmoorsState) then
        self.characterInfo.EttenmoorsState = true

        self.characterInfo.InstanceState = false
        if Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment() == 1 then
            self.characterInfo.InstanceVersion = 5
        elseif Turbine.Gameplay.LocalPlayer.GetInstance():GetAlignment() == 2 then
            self.characterInfo.InstanceVersion = 4
        end
    end

    if string.find(line, "resets") then
        -- DG
        if string.find(line, "Dâr Narbugud resets in") then
            self.characterInfo.InstanceState = true
            self:EnterNewInstance(line, 2)
        end

        -- Rift
        if string.find(line, "The Rift of Nûrz Ghâshu resets in") then
            self.characterInfo.InstanceState = true
            self:EnterNewInstance(line, 3)
        end

        -- Throne
        if string.find(line, "Throne of the Dread Terror resets in") then
            self.characterInfo.InstanceState = true
            self:EnterNewInstance(line, 8)
        end

        -- Filikul
        if string.find(line, "Filikul resets in") then
            self.characterInfo.InstanceState = true
            self:EnterNewInstance(line, 30)
        end

        -- Watcher
        if string.find(line, "The Vile Maw resets in") then
            self.characterInfo.InstanceState = true
            self:EnterNewInstance(line, 31)
        end

    end
end

function InstanceHub:CreateTables()

    --[[
        Instances
        Warg Pens of Dol Guldur
        Sammath Gûl

        Dark Delvings 21
        Fil Gashan 22
        Skumfil 23
        Forges of Khazad-dûm 24
        The Grand Stair 25
        The Sixteenth Hall 26

        Arboretum
        Barrows
        Fortress

        The Depths of Kidzul-kâlah
        The Harrowing of Morgul
        Eithel Gwaur, the Filth-well 
        Gath Daeroval, the Shadow-roost
        Gorthad Nûr, the Deep-barrow 

        Bâr Nírnaeth, the Houses of Lamentation 
        Ghashan-kútot, the Halls of Black Lore 
        The Fallen Kings 
        
        Shakalush, the Stair Battle

        Raids
        Barad Guldur
        Remmorchant, the Net of Darkness
        Amdân Dammul, the Bloody Threshold 
        Filikul
        The Vile Maw
    ]]

    self.questLines = {

        -- MM
        ["New Quest: Remmorchant, the Net of Darkness "] = 1,
        ["New Quest: Amdân Dammul, the Bloody Threshold "] = 7,
        ["New Quest: The Depths of Kidzul-kâlah "] = 9,
        ["New Quest: Shakalush, the Stair Battle "] = 10,
        ["New Quest: The Harrowing of Morgul"] = 11,
        ["New Quest: Bâr Nírnaeth, the Houses of Lamentation "] = 12,
        ["New Quest: Eithel Gwaur, the Filth-well "] = 13,
        ["New Quest: Gath Daeroval, the Shadow-roost "] = 14,
        ["New Quest: Ghashan-kútot, the Halls of Black Lore "] = 15,
        ["New Quest: Gorthad Nûr, the Deep-barrow "] = 16,
        ["New Quest: The Fallen Kings "] = 17,

        ["New Quest: Agoroth, the Narrowdelve "] = 32,
        ["New Quest: Woe of the Willow "] = 33,

        ["New Quest: Askâd-mazal, the Chamber of Shadows "] = 36,

        -- Mirkwood
        ["New Quest: Barad Guldur: The Deep Assault"] = 18,
        ["New Quest: Dark Deeds in Sammath Gûl"] = 19,
        ["New Quest: Strange Meat"] = 20,

        -- Moria
        ["New Quest: Challenge: Gurvand (Daily)"] = 21,
        ["New Quest: Challenge: General Talug (Daily)"] = 22,
        ["New Quest: Challenge: Hwandrin (Daily)"] = 23,
        ["New Quest: Challenge: Kranklûk (Daily)"] = 24,
        ["New Quest: Challenge: Igash (Daily)"] = 25,
        ["New Quest: Challenge: The Lost One (Daily)"] = 26,

        -- Garth Aragwen
        ["New Quest: Challenge: Grimbark"] = 27,
        ["New Quest: Challenge: Ivar"] = 28,
        ["New Quest: Challenge: Dúnlang"] = 29,   

        -- Filikul 30
        -- Vile maw 31

        -- Festival
        ["New Quest: Boss from the Vaults: Naruhel"] = 34,
        ["New Quest: Bee's Big Business"] = 35,

        -- ROots of fangorn
        ["New Quest: Instance: The Roots of Fangorn"] = 37,

        -- Fall of Khazad-Dum
        ["New Quest: The Fall of Khazad-dûm "] = 38,

        ["New Quest: Assault on Dhúrstrok "] = 39,
        ["New Quest: Den of Pughlak "] = 40,
        ["New Quest: Thikil-gundu, the Steel Keep"] = 41,
        ["New Quest: Caverns of Thrumfall "] = 42,
        ["New Quest: Glimmerdeep "] = 43,

        ["New Quest: Defenders of the Fortress"] = 44,
        ["New Quest: Uruk-necklaces"] = 45,

        --Annuminas
        ["New Quest: Glinghant Corrupted"] = 46,
        ["New Quest: End Dolvaethor's Menace"] = 47,
        ["New Quest: Balhest, Lord of Ost Elendil"] = 48,

        -- Barrows
        ["New Quest: Sambrog"] = 49,
        ["New Quest: Thadúr the Ravager"] = 50,
        ["New Quest: Gaerdring and Gaerthel"] = 51,
    
        -- Fornost
        ["New Quest: Raid: Spider Squasher"] = 52,
        ["New Quest: Raid: Giant-slayer"] = 53,
        ["New Quest: Raid: Drake-wrangler"] = 54,
        ["New Quest: Raid: Thorog Defeated"] = 55,

        ["New Quest: Adkhât-zahhar, the Houses of Rest "] = 56,
        
        ["New Quest: The Hiddenhoard of Abnankâra "] = 57,

        -- OD
        ["New Quest: Instance: The Lost Temple"] = 58,
        ["New Quest: Challenge: Toes in the Water"] = 59,
        ["New Quest: Challenge: The Power of Gortheron"] = 60,
        ["New Quest: Challenge: Face Your Fears"] = 61,
        ["New Quest: Challenge: Sâri-surma -- Survival"] = 62,
        ["New Quest: Challenge: The Unquenched Flame"] = 63,
        ["New Quest: Challenge: The Town of Stoneheight"] = 64,

        -- Erebor
        ["New Quest: Challenge: An Army from the East"] = 65,
        ["New Quest: Iorbar's Peak - Challenge"] = 66,
        ["New Quest: Challenge: Rapid Succession"] = 67,
        ["New Quest: The Battle for Erebor - Challenge"] = 68,
        ["New Quest: Challenge: The Bells of Dale"] = 69,
        ["New Quest: The Fires of Smaug - Challenge"] = 70,
        ["New Quest: Challenge: Webs of the Scuttledells"] = 71,
        
        -- Firework
        ["New Quest: A Flurry of Fireworks"] = 72,

        -- Before the Shadow
        ["New Quest: Doom of Caras Gelebren"] = 73,
        ["New Quest: Sarch Vorn, the Black Grave "] = 74,

        -- 
        ["New Quest: Sant Lhoer, the Poison Gardens "] = 75,
        ["New Quest: Sagroth, Lair of Vermin "] = 76,
        ["New Quest: Thaurisgar, the Vile Apothecary "] = 77,
        ["New Quest: Gwathrenost, the Witch-king's Citadel "] = 78,
        -- Highest 78

        ["New Quest: Fornost: Rhavameldir, the Wraith of Earth"] = 79,
        ["New Quest: Fornost: Brogadan, Wraith of Fire"] = 80,
        ["New Quest: Fornost: Megoriath, the Wraith of Water"] = 81,
        ["New Quest: Fornost: Remmenaeg, Wraith of Shadow"] = 82,

    }

    self.nameTable = {
        [0] = "All",
        [1] = "Remmorchant",
        [2] = "Dâr Narbugud",
        [3] = "The Rift of Nûrz Ghâshu",
        [4] = "Monster Player",
        [5] = "The Free Peoples",
        [6] = "The Ettenmoors",
        [7] = "Amdân Dammul",
        [8] = "Throne of the Dread Terror",
        [9] = "The Depths of Kidzul-kâlah",
        [10] = "Shakalush, the Stair Battle",
        [11] = "The Harrowing of Morgul",
        [12] = "Bâr Nírnaeth, the Houses of Lamentation",
        [13] = "Eithel Gwaur, the Filth-well",
        [14] = "Gath Daeroval, the Shadow-roost",
        [15] = "Ghashan-kútot, the Halls of Black Lore",
        [16] = "Gorthad Nûr, the Deep-barrow",
        [17] = "The Fallen Kings",
        [18] = "Barad Guldur",
        [19] = "Sammath Gûl",
        [20] = "Warg Pens of Dol Guldur",
        [21] = "Dark Delvings",
        [22] = "Fil Gashan",
        [23] = "Skumfil",
        [24] = "Forges of Khazad-dûm",
        [25] = "The Grand Stair",
        [26] = "The Sixteenth Hall",
        [27] = "Arboretum",
        [28] = "Barrows",
        [29] = "Fortress",
        [30] = "Filikul",
        [31] = "The Vile Maw",
        [32] = "Agoroth, the Narrowdelve",
        [33] = "Woe of the Willow",
        [34] = "Boss from the Vaults: Naruhel",
        [35] = "Bee's Big Business",
        [36] = "Askâd-mazal, the Chamber of Shadows",
        [37] = "Roots of Fangorn",
        [38] = "The Fall of Khazad-dûm",
        [39] = "Assault on Dhúrstrok",
        [40] = "Den of Pughlak",
        [41] = "Thikil-gundu",
        [42] = "Thrumfall",
        [43] = "Glimmerdeep",
        [44] = "Sâri-surma",
        [45] = "Sword-halls",
        [46] = "Glinghant",
        [47] = "Haudh Valandil",
        [48] = "Ost Elendil",
        [49] = "Sambrog",
        [50] = "Thadúr",
        [51] = "Great-barrow",
        [52] = "Spider Wing",
        [53] = "Giant Wing",
        [54] = "Drake Wing",
        [55] = "Thorog",
        [56] = "Adkhât-zahhar, the Houses of Rest",
        [57] = "The Hiddenhoard of Abnankâra",
        [58] = "Lost Temple",
        [59] = "Ost Dunhoth - Disease and Poison",
        [60] = "Ost Dunhoth - Gortheron",
        [61] = "Ost Dunhoth - Wound and Fear",
        [62] = "Sâri-surma",
        [63] = "The Northcotton Farm",
        [64] = "Stoneheight",
        [65] = "Flight to the Lonely Mountain",
        [66] = "Iorbar's Peak",
        [67] = "Seat of the Great Goblin",
        [68] = "The Battle for Erebor",
        [69] = "The Bells of Dale",
        [70] = "The Fires of Smaug",
        [71] = "Webs of Scuttledells",
        [72] = "A Flurry of Fireworks",
        [73] = "Doom of Caras Gelebren",
        [74] = "Sarch Vorn",
        [75] = "Sant Lhoer, the Poison Gardens",
        [76] = "Sagroth, Lair of Vermin",
        [77] = "Thaurisgar, the Vile Apothecary",
        [78] = "Gwathrenost, the Witch-king's Citadel",
        [79] = "Wrath of Earth",
        [80] = "Wrath of Fire",
        [81] = "Wrath of Water",
        [82] = "Wrath of Shadow",
    }

    self.delvingMissions = {
        ["New Quest: Mission: A Plague of Men"] = 1,
        ["New Quest: Mission: Brigands in Nímbarth"] = 2,
        ["New Quest: Mission: Caranost Command Post"] = 3,
        ["New Quest: Mission: Caranost Courtyard"] = 4,
        ["New Quest: Mission: Crown of Caranost"] = 5,
        ["New Quest: Mission: Down on the Farm"] = 6,
        ["New Quest: Mission: Haunting at Dol Ernil"] = 7,
        ["New Quest: Mission: Jorthkyn and Hounds"] = 8,
        ["New Quest: Mission: Siege of Sírlond"] = 9,
        ["New Quest: Mission: Steps of Caranost"] = 10,
        ["New Quest: Mission: Stirring of Spirits"] = 11,
        ["New Quest: Mission: The Bridge to Caranost"] = 12,
        ["New Quest: Mission: The Nettinglade"] = 13,
        ["New Quest: Mission: Siege-breaking"] = 14,
        ["New Quest: Mission: Tharbad Tower Courtyard"] = 15,
        ["New Quest: Mission: The Ruins of Amon Firn"] = 16,
        ["New Quest: Mission: The Tearful Tomb"] = 17,
        ["New Quest: Mission: Before They Bond"] = 18,
        ["New Quest: Mission: The Ruins of Gaervarad"] = 19,
        ["New Quest: Mission: Signs and Sigils"] = 20,
        ["New Quest: Mission: A Day on the Farm"] = 21,
        ["New Quest: Mission: Dangerous Deliveries"] = 22,
        ["New Quest: Mission: Not as Many Heads"] = 23,
        ["New Quest: Mission: Leadership Direction"] = 24,
        ["New Quest: Mission: The Jaws of Death"] = 25,
        ["New Quest: Mission: Aberrations of Nature"] = 26,
        ["New Quest: Mission: Infiltration"] = 27,
        ["New Quest: Mission: Stirring on Sarn Ford"] = 28,
        ["New Quest: Mission: Patrol the Township"] = 29,
        ["New Quest: Mission: Terrible Turtle Trouble"] = 30,
        ["New Quest: Mission: Mother of Wolves"] = 31,
        ["New Quest: Mission: Secrets and Plans"] = 32,
        ["New Quest: Mission: Among the Watch-towers"] = 33,
        ["New Quest: Mission: Into the Pits"] = 34,
        ["New Quest: Mission: Bite the Hand"] = 35,
        ["New Quest: Mission: Sifting Through Time"] = 36,
        ["New Quest: Mission: The Hand, Out!"] = 37,
        ["New Quest: Mission: The Tombs of Gond Orchal"] = 38,
        ["New Quest: Mission: Shadows in the Greyflood"] = 39,
        ["New Quest: Mission: Troubled Waters"] = 40,
        ["New Quest: Mission: The Accursed Flame"] = 41,
        ["New Quest: Mission: Tharbad Tower Ascent"] = 42,

        ["All"] = 999,
    }

    self.delvingMissionsNames = {
        [1] = "A Plague of Men",
        [2] = "Brigands in Nímbarth",
        [3] = "Caranost Command Post",
        [4] = "Caranost Courtyard",
        [5] = "Crown of Caranost",
        [6] = "Down on the Farm",
        [7] = "Haunting at Dol Ernil",
        [8] = "Jorthkyn and Hounds",
        [9] = "Siege of Sírlond",
        [10] = "Steps of Caranost",
        [11] = "Stirring of Spirits",
        [12] = "The Bridge to Caranost",
        [13] = "The Nettinglade",
        [14] = "Siege-breaking",
        [15] = "Tharbad Tower Courtyard",
        [16] = "The Ruins of Amon Firn",
        [17] = "The Tearful Tomb",
        [18] = "Before They Bond",
        [19] = "The Ruins of Gaervarad",
        [20] = "Signs and Sigils",
        [21] = "A Day on the Farm",
        [22] = "Dangerous Deliveries",
        [23] = "Not as Many Heads",
        [24] = "Leadership Direction",
        [25] = "The Jaws of Death",
        [26] = "Aberrations of Nature",
        [27] = "Infiltration",
        [28] = "Stirring on Sarn Ford",
        [29] = "Patrol the Township",
        [30] = "Terrible Turtle Trouble",
        [31] = "Mother of Wolves",
        [32] = "Secrets and Plans",
        [33] = "Among the Watch-towers",
        [34] = "Into the Pits",
        [35] = "Bite the Hand",
        [36] = "Sifting Through Time",
        [37] = "The Hand, Out!",
        [38] = "The Tombs of Gond Orchal",
        [39] = "Shadows in the Greyflood",
        [40] = "Troubled Waters",
        [41] = "The Accursed Flame",
        [42] = "Tharbad Tower Ascent",


        [999] = "All",
    }
end

function InstanceHub:Unload()
    SaveTable(Turbine.DataScope.Character, "PrimeUtilsInstances", self.characterInfo)
    SaveTable(Turbine.DataScope.Character, "PrimeUtilsTEQData", self.TEQStats)
end

function InstanceHub:GetDungeonNameTable()
    return self.nameTable
end

function InstanceHub:GetDelvingTable()
    return self.delvingMissionsNames
end

function InstanceHub:GetFightNames()
    local nameTable = {}
    for instanceID, bossTable in pairs(self:GetRaidInfo()) do
        for bossStartTrigger, bossInfo in pairs(bossTable) do
            for startString, bossInfoTable in pairs(bossInfo) do
                nameTable[bossInfoTable[4]] = true 
            end
        end
    end
    return nameTable
end

function InstanceHub:UpdateDelvingData(completed)
    -- Setup variables
    local mission = Settings.MissionQuest
    local delvingTier = Settings.DelvingTier
    local startTime = Settings.MissionStartTime
    local endTime = Turbine.Engine.GetGameTime()
    local time = endTime - startTime
    local timeString = FormatTimeStringNoMilli(time)

    -- Null checks
    if not(self.TEQStats[mission]) then
        self.TEQStats[mission] = {
        }
    end

    if not(self.TEQStats[mission][delvingTier]) then
        self.TEQStats[mission][delvingTier] = {
            ["Completions"] = 0,
            ["TotalTime"] = 0,
            ["Fails"] = 0,
            ["Fastest"] = -1,
            ["Slowest"] = -1,
            ["Tries"] = 0,
        }
    end

    -- Completion types
    if completed then
        self.TEQStats[mission][delvingTier]["Completions"] = self.TEQStats[mission][delvingTier]["Completions"] + 1

        --Chat Message completion + time
        ChatOutput(string.format("%sYou have completed the delving: %s in %s. You have completed this mission on this delving tier %d times. %s", "<rgb=#0008CFF>", mission, timeString, self.TEQStats[mission][delvingTier]["Completions"], "</rgb>"))

        -- General times
        self.TEQStats[mission][delvingTier]["Tries"] = self.TEQStats[mission][delvingTier]["Tries"] + 1
        self.TEQStats[mission][delvingTier]["TotalTime"] = self.TEQStats[mission][delvingTier]["TotalTime"] + time

        local averageTimeString = FormatTimeStringNoMilli(self.TEQStats[mission][delvingTier]["TotalTime"] / self.TEQStats[mission][delvingTier]["Completions"])

        -- Fastest time
        if self.TEQStats[mission][delvingTier]["Fastest"] == -1 or self.TEQStats[mission][delvingTier]["Fastest"] > time then
            self.TEQStats[mission][delvingTier]["Fastest"] = time

            -- Fastest time update
            ChatOutput(string.format("%sYour time %s is your fastest completion for this delving on this tier! Your average completion time is %s.%s", "<rgb=#0008CFF>", timeString, averageTimeString, "</rgb>"))
        else
            -- Average time update
            ChatOutput(string.format("%sYour average completion time for this mission on this delving tier is %s.%s", "<rgb=#0008CFF>", averageTimeString, "</rgb>"))
        end

        -- Slowest time
        if self.TEQStats[mission][delvingTier]["Slowest"] == -1 or self.TEQStats[mission][delvingTier]["Slowest"] < time then
            self.TEQStats[mission][delvingTier]["Slowest"] = time
        end
    else
        self.TEQStats[mission][delvingTier]["Fails"] = self.TEQStats[mission][delvingTier]["Fails"] + 1

        -- Chat message failure
        ChatOutput(string.format("%sTotal time for this undertaking was %s.%s", "<rgb=#0008CFF>", timeString, "</rgb>"))
    end
end
